/****************************************************************************
 *
 * Palm Universal Wireless Keyboard(TM) add-on driver for linux-Xwindow. v0.3.3
 * Copyright GPL Licence. 
 * 2005.05.14 lampo1 <lampo1@paran.com>
 *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XTest.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>

struct keymapping {
    unsigned char hwcode;       /* Input keycode from IR Keyboard to Linux
				   IrDA port( = this program) */
    unsigned int xcode_nml;     /* Output keycode from this program to X
				   server according to xmodmap */ 
    unsigned int xcode_lfn;     /* xcode with Left-Fn key */
};

/* This is default keymapping of debian woody */ 
struct keymapping KeyTable[64] = {
    {0x0e, 49, 0},       /* Grave grave:asciitilde none */
    {0x16, 10, 0},       /* 1 1:exclam none */
    {0x1e, 11, 0},       /* 2 2:at none */
    {0x26, 12, 0},       /* 3 3:numbersign none */
    {0x25, 13, 0},       /* 4 4:dollar none */
    {0x2e, 14, 0},       /* 5 5:percent none */
    {0x36, 15, 0},       /* 6 6:asciicircum none */
    {0x3d, 16, 0},       /* 7 7:ampersand none */
    {0x3e, 17, 0},       /* 8 8:asterisk none */
    {0x46, 18, 0},       /* 9 9:parenleft none */
    {0x45, 19, 0},       /* 0 0:parenright none */
    {0x4e, 20, 0},       /* Minus minus:underscore none */
    {0x55, 21, 0},       /* Equal equal:plus none */
    {0x66, 22, 0},       /* BackSp BackSpace:Terminate_Server none */
    {0x0d, 23, 0},       /* Tab Tab:ISO_Left_Tab none */
    {0x15, 24, 0},       /* Q q:Q none */
    {0x1d, 25, 0},       /* W w:W none */
    {0x24, 26, 0},       /* E e:E none */
    {0x2d, 27, 0},       /* R r:R none */ 
    {0x2c, 28, 0},       /* T t:T none */
    {0x35, 29, 0},       /* Y y:Y none */
    {0x3c, 30, 0},       /* U u:U none */
    {0x43, 31, 0},       /* I i:I none */
    {0x44, 32, 0},       /* O o:O none */
    {0x4d, 33, 0},       /* P p:P none */
    {0x54, 34, 0},       /* BracketLeft bracketleft:braceleft none */
    {0x5b, 35, 0},       /* BracketRight bracketright:braceright none */
    {0x5d, 51, 0},       /* BackSlash backslash:bar none */
    {0x58, 66, 0},       /* CapsLk Caps_Lock:none none */
    {0x1c, 38, 0},       /* A a:A none */
    {0x1b, 39, 0},       /* S s:S none */
    {0x23, 40, 0},       /* D d:D none */
    {0x2b, 41, 0},       /* F f:F none */
    {0x34, 42, 0},       /* G g:G none */
    {0x33, 43, 0},       /* H h:H none */
    {0x3b, 44, 0},       /* J j:J none */
    {0x42, 45, 0},       /* K k:K none */
    {0x4b, 46, 0},       /* L l:L none */
    {0x4c, 47, 0},       /* Semicolon semicolon:colon none */
    {0x52, 48, 0},       /* Apostrophe apostrophe:quotedbl none */
    {0x5a, 36, 0},       /* Enter Return:none none */
    {0x12, 50, 0},       /* Shift-L Shift_L:none none */
    {0x1a, 52, 0},       /* Z z:Z none */
    {0x22, 53, 0},       /* X x:X none */
    {0x21, 54, 0},       /* C c:C none */
    {0x2a, 55, 0},       /* V v:V none */
    {0x32, 56, 0},       /* B b:B none */
    {0x31, 57, 0},       /* N n:N none */
    {0x3a, 58, 0},       /* M n:M none */
    {0x48, 59, 0},       /* Comma comma:less none */
    {0x49, 60, 0},       /* Period period:greater none */
    {0x28, 98, 99},      /* Up Up:none Prior */
    {0x59, 62, 0},       /* Shift-R Shift_R:none none */
    {0x4a, 61, 0},       /* Slash slash:question none */
    {0x14, 37, 0},       /* Ctra-L Control_L:none none */
    {0x02, 0, 0},        /* Left-Fn none:none none */
    {0x03, 9, 0},        /* Cmd Escape:none none */
    {0x11, 0, 0},        /* Cancel none:none none */
    {0x29, 65, 0},       /* Space space:none none */
    {0x30, 113, 0},      /* Alt_Gr Alt_R:Meta_R none */
    {0x5e, 100, 97},     /* Left Left:none Home */
    {0x60, 104, 105},    /* Down Down:none Next */
    {0x2f, 102, 103},    /* Right Right:none End */
    {0x1f, 107, 0}       /* Del Delete:none none */ 
};

int fd;
unsigned int auto_off_time = 300;    /* seconds */
unsigned int typematic_rate = 1;     /* n times of unit cycle */
unsigned int typematic_delay = 3;    /* n times of unit cycle */
unsigned int num_rptcode = 0;
unsigned int lfn_FLAG = 0;
Display *d;

void alarm_inthandler(int fd) {
    close(fd);
    exit(EXIT_SUCCESS);
}
    
void read_rcfile() {
    int i;
    FILE *fp;
    char dmy[255];
    
    fp = fopen("/etc/uwkrc", "r");
    if ( fp == NULL ) {
	fprintf(stderr,"Cannot open /etc/uwkrc.\n");
	fprintf(stderr,"Internal key definition is applied.\n");
	return;
    }

    while ( (fgets(dmy, sizeof(dmy), fp)) != NULL ) {

	if ( dmy[0] == '#' ) continue;
	else if ( dmy[0] == 'd' ) {
	    sscanf(dmy, "%*s %d", &typematic_delay);
	} else if ( dmy[0] == 'r' ) {
	    sscanf(dmy, "%*s %d", &typematic_rate);
	} else if ( dmy[0] == 'a' ) {
	    sscanf(dmy, "%*s %d", &auto_off_time);
	} else if ( isdigit( dmy[0] ) != 0 ) {
	    for ( i=0; i <= 63; i++) {
		sscanf(dmy,"%x %ud %ud", 
		       &KeyTable[i].hwcode, 
		       &KeyTable[i].xcode_nml, 
		       &KeyTable[i].xcode_lfn);
		fgets(dmy, sizeof(dmy), fp);
	    } 
	}
    }
    fclose(fp);
}

void translate( unsigned char ch) {

    int  i;
    unsigned char msb;   /* 0 : key pressed */
                         /* 1 means key released. */

    msb =  (ch & 128) >> 7 ;   
    ch &= 127 ; 

    for (i = 0; i <= 63 ; i++) {

	if ( ch == KeyTable[i].hwcode ) {
	    if ( ch == 0x02 ) {         /* Left-Fn */
		lfn_FLAG = (1^msb);
		break;
	    } else if ( lfn_FLAG == 1 ) {
		if ( ch == 0x1f ) {          /* LeftFn-Del */
		    close(fd);
		    exit(EXIT_SUCCESS);
		}
		XTestFakeKeyEvent(d, KeyTable[i].xcode_lfn, (1^msb), 0);
		XFlush(d);
		break;		    
	    } else {
		XTestFakeKeyEvent(d, KeyTable[i].xcode_nml, (1^msb), 0);
		XFlush(d);
		break;
	    }
	}
    }
}
	
int main(int argc, char *argv[]) {

    unsigned char ch, hwcode;
    int nread;
    char *device = "/dev/ttyS1";

    struct termios oldtio, newtio;
    struct sigaction newhandler;
    sigset_t blocked;

    /* signal install */
    newhandler.sa_handler = alarm_inthandler;
    sigemptyset(&blocked);
    sigaddset(&blocked, SIGALRM);
    newhandler.sa_mask = blocked;
    if ( sigaction(SIGALRM, &newhandler, NULL) == -1 ) {
	perror("sigaction install error\n");
	exit(EXIT_FAILURE);
    }

    /* Opening irda port device */
    if ( argc >= 2 ) device = argv[1]; 
    fd = open(device, O_RDONLY | O_NOCTTY );
    if (fd == -1)  {
    	fprintf(stderr, "Cannot open %s\n", device);
    	exit(EXIT_FAILURE);
    }

    read_rcfile();

    /* Serial port setting to noncanonical, no echo, ...*/
    tcgetattr(fd, &oldtio); 
    bzero(&newtio, sizeof(newtio));
    newtio.c_cflag = B9600 | CRTSCTS | CS8 | CLOCAL | CREAD;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;
    newtio.c_lflag = 0;
    newtio.c_cc[VTIME]    = 0;   /* inter-character timer unused */
    newtio.c_cc[VMIN]     = 1;   /* blocking read until 1 chars received */
    tcflush(fd, TCIFLUSH);
    tcsetattr(fd,TCSANOW,&newtio);
               
    d = XOpenDisplay(NULL);
    
    while (1) {
	alarm( auto_off_time );

	nread = read(fd, &ch, 1);
        if ( ( nread > 0 ) && ( ch == 0xff ) ) {
      
	    nread = read(fd, &ch, 1);
	    if ( (nread > 0) && ( ch == 0xc0 ) ) {
	    
		nread = read(fd, &hwcode, 1);
		read(fd, &ch, 1);
		read(fd, &ch, 1); 
		nread = read(fd, &ch, 1);

		if ( ch == 0xc1 ) translate(hwcode);		
	    }
	} else if ( nread == -1 ) {
	    fprintf(stderr,"IrDA Port reading error.\n");
	    fprintf(stderr,"Recommend removing other irda modules.\n");
	    exit(EXIT_FAILURE);
	}         
    }  
}
